/*
 * time.c
 *
 * Copyright 2002, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 */


#include"types.h"
#include"lib.h"
#include"errno.h"
#include"interrupt.h"
#include"proc.h"
#include"sp.h"
#include"mp.h"
#include"lock.h"
#include"mm.h"
#include"time.h"


enum{
	MAX_TIMER=4,			/* ݻޡ */

	/* Timer type */
	TYP_INTERVAL=1,			/* Interval timer */
	TYP_REAL=2,				/* Realclock timer */
};


typedef struct TIMER_LINK{
	struct TIMER_LINK *next;
	struct TIMER_LINK *prev;
	int count;					/* Number of used real timer. */
	int lock_gate;				/* ԥåȡ */
}TIMER_LINK;


int clock_1micro;		/* 1micros */
int clock_1m;			/* 1ms */


/*
 * PUBLIC
 * TIMER_LINK¤Υɥ쥹PROCɥ쥹롣
 */
static inline PROC *timerLinkToProc(TIMER_LINK *p)
{
	return (PROC*)((uint)p-(uint)&((PROC*)0)->wait_next);
}


/***************************************************************************************************
 *
 * 8254 interval timer
 *
 ***************************************************************************************************/

static uint tasking_count=				/* 󥰥ץåѥޡ */
	PIT_HZ/(1000/TASK_TIME);
static TIMER_LINK intervalTimer=		/* Interval timer link queue */
	{&intervalTimer,&intervalTimer,0,0};
static int interval_current_count=0;	/* Interval timer current count */


/*
 * GLOBAL
 * 8254 interval timer handler
 */
static int interval_timer_handler()
{
	PROC *proc;
	TIMER_LINK *p;
	TIMER *timer;

/****************************************************************/
	printk("Interrupt IRQ0\n");
/*****************************************************************/
	enter_spinlock(&intervalTimer.lock_gate);
	{
		/* ॢȤΥץ褵 */
		for(p=intervalTimer.next;;p=p->next)
		{
			proc=timerLinkToProc(p);
			timer=proc->timer_struct;
			if(timer->timer_count!=interval_current_count)break;

			timer->timer_count=0;
			p->prev->next=p->next;
			p->next->prev=p->prev;
			add_to_schedule(proc,TASK_DEVICE_WAIT);

			--intervalTimer.count;
		}

		/* No timer */
		if(intervalTimer.count==0)
		{
			if(MFPS_addres==0)
			{
				tasking_count=PIT_HZ/(1000/TASK_TIME);
				init_sp_task();
			}
		}
		else
		{
			/* ƥץΥȤ򸺤餹 */
			for(p=intervalTimer.next;p!=&intervalTimer;p=p->next)
			{
				proc=timerLinkToProc(p);
				timer=proc->timer_struct;
				timer->timer_count-=interval_current_count;
			}

			proc=timerLinkToProc(intervalTimer.prev);
			timer=proc->timer_struct;
			if(timer->timer_count>0x10000)interval_current_count=0x10000;
			else interval_current_count=timer->timer_count;

			if(MFPS_addres==0)
			{
				if((tasking_count-=interval_current_count)==0)
					tasking_count=PIT_HZ/(1000/TASK_TIME);

				if(tasking_count<interval_current_count)interval_current_count=tasking_count;
			}

			/* ޡꤹ */
			outb(PIT_CNR0,interval_current_count&0xff);
			outb(PIT_CNR0,interval_current_count>>8);
		}
	}
	exit_spinlock(&intervalTimer.lock_gate);

	return 1;
}


/*
 * PUBLIC
 * Set interval timer
 * parameters : time(msñ)
 * rturn : Task switch flag or error number
 */
static int set_interval_timer(int time)
{
	uint count;
	PROC *proc,*pr;
	TIMER_LINK *p,*q;
	TIMER *timer,*timer2;


	if(intervalTimer.count==MAX_TIMER)return -ENOTIMER;

	time=PIT_HZ*time/1000;				/* åѴ */

	proc=get_current_task();
	p=(TIMER_LINK*)&proc->wait_next;
	timer=proc->timer_struct;

	enter_spinlock(&intervalTimer.lock_gate);
	{
		if(intervalTimer.count==0)
		{
			timer->timer_count=time;
			p->next=p->prev=&intervalTimer;
			intervalTimer.next=intervalTimer.prev=p;

			if(MFPS_addres==0)
			{
				irq_entry[IRQ0]=interval_timer_handler;
				outb(PIT_CTR,0x30);					/* One shot timer(mode 0) */
				if(time>PIT_HZ/(1000/TASK_TIME))
					time=PIT_HZ/(1000/TASK_TIME);
			}

			/* ޡ */
			if(time>0x10000)interval_current_count=0x10000;
			else interval_current_count=time;
			outb(PIT_CNR0,interval_current_count&0xff);
			outb(PIT_CNR0,interval_current_count>>8);
		}
		else
		{
			count=inb(PIT_CNR0);
			count+=inb(PIT_CNR0)<<8;
			timer->timer_count=time+count;

			/* ޡ󥯤³ */
			for(q=intervalTimer.next;q!=&intervalTimer;q=q->next)
			{
				pr=timerLinkToProc(q);
				timer2=pr->timer_struct;
				if(timer2->timer_count>=timer->timer_count)break;
			}
			p->next=q;
			p->prev=q->prev;
			p->prev->next=q->prev=p;

			if(timer->timer_count<interval_current_count)
			{
				/* ޡ */
				interval_current_count=timer->timer_count;
				outb(PIT_CNR0,time&0xff);
				outb(PIT_CNR0,time>>8);
			}
		}
	}
	exit_spinlock(&intervalTimer.lock_gate);

	++intervalTimer.count;
	timer->timer_type=TYP_INTERVAL;

	return 0;
}


/*
 * PUBLIC
 * Delete interval timer from link
 * parameters : Process number
 */
static void del_interval_timer(PROC *proc)
{
	TIMER_LINK *p;
	TIMER *timer;


	p=(TIMER_LINK*)&proc->wait_next;
	timer=proc->timer_struct;

	enter_spinlock(&intervalTimer.lock_gate);
	{
		if(timer->timer_count!=0)
		{
			p->prev->next=p->next;
			p->next->prev=p->prev;
			p->next=p->prev=p;
			timer->timer_count=0;
		}

		if(--intervalTimer.count==0)
		{
			if(MFPS_addres==0)
			{
				tasking_count=PIT_HZ/(1000/TASK_TIME);
				init_sp_task();
			}
			else outb(PIT_CTR,0x30);		/* ޡȥå */
		}
	}
	exit_spinlock(&intervalTimer.lock_gate);
}


/***************************************************************************************************
 *
 * MC146818 realclock timer
 *
 ***************************************************************************************************/

enum{
	REAL_TIMER_RATE=500,	/* Real-clock timer rate(ms) */
};


static TIMER_LINK realTimer=		/* Real timer link queue */
	{&realTimer,&realTimer,0,0};
static uint utc;					/* current UTC(number of seconds since 00:00:00.1.1.1970). */


/*
 * PRIVATE
 * ʿʿѴ롣
 */
static inline uchar bcd_to_bin(uchar value)
{
	return value-6*(value>>4);
}


/*
 * PUBLIC
 * Set realclock timer
 * parameters : time(msñ)
 * rturn : 0 or error number
 */
static int set_real_timer(int time)
{
	PROC *proc;
	TIMER_LINK *p;
	TIMER *timer;


	if(realTimer.count==MAX_TIMER)return -ENOTIMER;

	proc=cputask[get_current_cpu()].current_task;
	p=(TIMER_LINK*)&proc->wait_next;
	timer=proc->timer_struct;
	timer->timer_count=time;

	/* 󥯤³ */
	enter_spinlock(&realTimer.lock_gate);
	{
		p->next=realTimer.next;
		p->prev=&realTimer;
		realTimer.next->prev=p;
		realTimer.next=p;
	}
	exit_spinlock(&realTimer.lock_gate);

	++realTimer.count;
	timer->timer_type=TYP_REAL;

	return 0;
}


/*
 * PUBLIC
 * Delete realclock timer
 * prameters : process
 */
static void del_real_timer(PROC *proc)
{
	TIMER_LINK *p;
	TIMER *timer;	


	p=(TIMER_LINK*)&proc->wait_next;
	timer=proc->timer_struct;

	enter_spinlock(&realTimer.lock_gate);
	{
		if(timer->timer_count!=0)
		{
			p->prev->next=p->next;
			p->next->prev=p->prev;
			p->next=p->prev=p;
			timer->timer_count=0;
			--realTimer.count;
		}
	}
	exit_spinlock(&realTimer.lock_gate);
}


/*
 * GLOBAL
 * MC146818 real timer handler
 * return : ץåʤ
 */
static int real_timer_handler()
{
	static int half=0;
	PROC *proc;
	TIMER_LINK *p;
	TIMER *timer;


	/* Status register C ɤǳߤꥻåȤ롣 */
	read_cmos(CMOS_STRC);

	/* ॢȥץĴ٤ */
	enter_spinlock(&realTimer.lock_gate);
	{
		for(p=realTimer.next;p!=&realTimer;p=p->next)
		{
			proc=timerLinkToProc(p);
			timer=proc->timer_struct;
			if((timer->timer_count-=REAL_TIMER_RATE)<=0)		/* ॢ */
			{
				p->prev->next=p->next;
				p->next->prev=p->prev;
				timer->timer_count=0;
				add_to_schedule(proc,TASK_DEVICE_WAIT);
				--realTimer.count;
			}
		}
	}
	exit_spinlock(&realTimer.lock_gate);

	/* UTC򹹿롣 */
	half^=1;				/* 500msȹ */
	utc+=half;

	return 0;
}


/*
 * GLOBAL
 * get current time from cmos
 * parameters : time structure
 */
static void get_time_from_cmos()
{
	static ushort sumMonthDay[13]=				/* ιס */
		{0,0,31,59,90,120,151,181,212,243,273,304,334};

	struct TIME{
		uchar second;		/* seconds */
		uchar minute;		/* minutes */
		uchar hour;			/* hours */
		uchar day;			/* days */
		uchar month;		/* months */
		ushort year;		/* years */
	}time;
	int days;


	do
	{
		time.second=bcd_to_bin(read_cmos(CMOS_SCD));
		time.minute=bcd_to_bin(read_cmos(CMOS_MNT));
		time.hour=bcd_to_bin(read_cmos(CMOS_HOR));
		time.day=bcd_to_bin(read_cmos(CMOS_DAY));
		time.month=bcd_to_bin(read_cmos(CMOS_MTH));
		time.year=bcd_to_bin(read_cmos(CMOS_YER));
	}while(bcd_to_bin(read_cmos(CMOS_SCD))<time.second);

	time.year+=(time.year<70)?2000:1900;

	/* UTCη׻ */
	days=(time.year-1970)*365;
	days+=(time.year-1-1970)/4;
	days+=sumMonthDay[time.month]+time.day-1;
	if((time.year%4==0)&&(time.month>2))++days;
	utc=((days*24+time.hour)*60+time.minute)*60+time.second;
	utc+=LOCAL_TIME_TOKYO;								/* 륿֤Ĵ */
}


/***************************************************************************************************
 *
 * misc functions
 *
 ***************************************************************************************************/

/*
 * GLOBAL
 * cpuå¬
 * ֤ : å
 */
int count_cpu_clock()
{
	uint cpu_clock;
	int low,high;
	long long r_beg,r_end;
	double time;


 	outb(PIT_CTR,0x30);	/* CLK0,LSB.MSB,⡼0,binary */
	outb(PIT_CNR0,0);	/* LSB */
	outb(PIT_CNR0,0);	/* MSB */

 	r_beg=rdtsc();		/* clock counter */

	for(;;)if(rdtsc()-r_beg>0x400000)break;		/* ٱ */

	r_end=rdtsc();

	low=inb(PIT_CNR0);
	high=inb(PIT_CNR0);

	time=(0xffff-low-(high<<8))/(double)PIT_HZ;
	cpu_clock=(r_end-r_beg)/time;

 	clock_1micro=cpu_clock/1000000+1;		/* 1micros */
 	clock_1m=cpu_clock/1000+1;				/* 1ms */

	return cpu_clock;
}


/*
 * GLOBAL
 * Set timer
 * ա
 *  δؿɬ del_from_schedule() θǻȤȡ⤷˥ॢȤ
 *  ץʤ롣
 * parameters : time(msñ)
 * rturn : Task switch flag or error number
 */
int set_timer(int time)
{
	if(time==0)return 0;

	if(time<REAL_TIMER_RATE)return set_interval_timer(time);
	else return set_real_timer(time);
}


/*
 * GLOBAL
 * Delete timer from timer link.
 * parameters : Process number
 */
void del_timer(PROC *proc)
{
	TIMER *timer=proc->timer_struct;
	
	
	if(timer->timer_type==TYP_INTERVAL)del_interval_timer(proc);
	if(timer->timer_type==TYP_REAL)del_real_timer(proc);
}


/*
 * GLOBAL
 * set time struct.
 * return : time struct address or error=NULL
 */
void *setTimerStruct()
{
	TIMER *timer;


	if((timer=kmalloc(sizeof(TIMER)))==NULL)return NULL;
	memset(timer,0,sizeof(TIMER));

	return timer;
}


/*
 * GLOBAL
 * release time struct.
 */
void releaseTimer(TIMER *timer)
{
	if(timer->timer_count!=0)del_timer(get_current_task());
	kfree(timer);
}
 
 
/***************************************************************************************************
 *
 * < ƥॳ >
 *
 ***************************************************************************************************/

/*
 * get Coordinated Universal Time(number of seconds since 00:00:00.1.1.1970)
 */
int sys_time()
{
	return (int)utc;
}


/***************************************************************************************************
 *
 * 
 *
 ***************************************************************************************************/

/*
 * GLOBAL
 * Init time
 */
void init_time()
{
	/* ޥץå */
	if(MFPS_addres)
	{
		irq_entry[IRQ0]=interval_timer_handler;
		outb(PIT_CTR,0x30);						/* One shot timer(mode 0). */
		release_irq_mask(0);
	}

	/* Init MC146818 real time clock */
	irq_entry[IRQ8]=real_timer_handler;
	get_time_from_cmos();				/* get current time from cmos. */
	write_cmos(CMOS_STRA,0x2f);			/* time bese 0x010(default)|ޡ졼65536/(20xf)=2HZ(500ms) */
	read_cmos(CMOS_STRC);				/* Status register C ɤǳߤꥻåȤ롣 */
	write_cmos(CMOS_STRB,0x42);			/* 24hour,ߥ */
	release_irq_mask(8);
}
/*************************************************************************************/
void test_time()
{

}
/**************************************************************************************/
